Imports System

Namespace MatrixInverse
    Friend Class MatrixInverseProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin matrix inverse using Crout LU decomp demo " & vbCrLf)

            Dim m()() As Double = MatrixCreate(4, 4)
            m(0)(0) = 3.0
            m(0)(1) = 7.0
            m(0)(2) = 2.0
            m(0)(3) = 5.0
            m(1)(0) = 1.0
            m(1)(1) = 8.0
            m(1)(2) = 4.0
            m(1)(3) = 2.0
            m(2)(0) = 2.0
            m(2)(1) = 1.0
            m(2)(2) = 9.0
            m(2)(3) = 3.0
            m(3)(0) = 5.0
            m(3)(1) = 4.0
            m(3)(2) = 7.0
            m(3)(3) = 1.0

            Console.WriteLine("Original matrix m is ")
            Console.WriteLine(MatrixAsString(m))

            Dim d As Double = MatrixDeterminant(m)
            If Math.Abs(d) < 0.00001 Then
                Console.WriteLine("Matrix has no inverse")
            End If

            Dim inv()() As Double = MatrixInverse(m)

            Console.WriteLine("Inverse matrix inv is ")
            Console.WriteLine(MatrixAsString(inv))

            Dim prod()() As Double = MatrixProduct(m, inv)
            Console.WriteLine("The product of m * inv is ")
            Console.WriteLine(MatrixAsString(prod))

            Console.WriteLine("========== " & vbCrLf)

            Dim lum()() As Double = Nothing
            Dim perm() As Integer = Nothing
            Dim toggle As Integer = MatrixDecompose(m, lum, perm)
            Console.WriteLine("The combined lower-upper decomposition of m is")
            Console.WriteLine(MatrixAsString(lum))

            Dim lower()() As Double = ExtractLower(lum)
            Dim upper()() As Double = ExtractUpper(lum)

            Console.WriteLine("The lower part of LUM is")
            Console.WriteLine(MatrixAsString(lower))

            Console.WriteLine("The upper part of LUM is")
            Console.WriteLine(MatrixAsString(upper))

            Console.WriteLine("The perm[] array is")
            ShowVector(perm)

            Dim lowTimesUp()() As Double = MatrixProduct(lower, upper)
            Console.WriteLine("The product of lower * upper is ")
            Console.WriteLine(MatrixAsString(lowTimesUp))

            Console.WriteLine(vbCrLf & "End matrix inverse demo " & vbCrLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Sub ShowVector(ByVal vector() As Integer)
            Console.Write("   ")
            For i = 0 To vector.Length - 1
                Console.Write(vector(i) & " ")
            Next i
            Console.WriteLine(vbCrLf)
        End Sub

        Private Shared Function MatrixInverse(ByVal matrix()() As Double) As Double()()
            ' assumes determinant is not 0
            ' that is, the matrix does have an inverse
            Dim n As Integer = matrix.Length
            Dim result()() As Double = MatrixCreate(n, n) ' make a copy of matrix
            For i = 0 To n - 1
                For j = 0 To n - 1
                    result(i)(j) = matrix(i)(j)
                Next j
            Next i

            Dim lum()() As Double = Nothing ' combined lower & upper
            Dim perm() As Integer = Nothing
            Dim toggle As Integer
            toggle = MatrixDecompose(matrix, lum, perm)

            Dim b(n - 1) As Double
            For i = 0 To n - 1
                For j = 0 To n - 1
                    If i = perm(j) Then
                        b(j) = 1.0
                    Else
                        b(j) = 0.0
                    End If
                Next j

                Dim x() As Double = Helper(lum, b)
                For j = 0 To n - 1
                    result(j)(i) = x(j)
                Next j
            Next i
            Return result
        End Function ' MatrixInverse

        Private Shared Function MatrixDecompose(ByVal m()() As Double, <System.Runtime.InteropServices.Out> ByRef lum()() As Double, <System.Runtime.InteropServices.Out> ByRef perm() As Integer) As Integer
            ' Crout's LU decomposition for matrix determinant and inverse
            ' stores combined lower & upper in lum][
            ' stores row permuations into perm[]
            ' returns +1 or -1 according to even or odd number of row permutations
            ' lower gets dummy 1.0s on diagonal (0.0s above)
            ' upper gets lum values on diagonal (0.0s below)

            Dim toggle As Integer = +1 ' even (+1) or odd (-1) row permutatuions
            Dim n As Integer = m.Length

            ' make a copy of m][ into result lu][
            lum = MatrixCreate(n, n)
            For i = 0 To n - 1
                For j = 0 To n - 1
                    lum(i)(j) = m(i)(j)
                Next j
            Next i

            ' make perm[]
            perm = New Integer(n - 1) {}
            For i = 0 To n - 1
                perm(i) = i
            Next i

            For j = 0 To n - 2 ' process by column. note n-1
                Dim max As Double = Math.Abs(lum(j)(j))
                Dim piv As Integer = j

                For i = j + 1 To n - 1 ' find pivot index
                    Dim xij As Double = Math.Abs(lum(i)(j))
                    If xij > max Then
                        max = xij
                        piv = i
                    End If
                Next i ' i

                If piv <> j Then
                    Dim tmp() As Double = lum(piv) ' swap rows j, piv
                    lum(piv) = lum(j)
                    lum(j) = tmp

                    Dim t As Integer = perm(piv) ' swap perm elements
                    perm(piv) = perm(j)
                    perm(j) = t

                    toggle = -toggle
                End If

                Dim xjj As Double = lum(j)(j)
                If xjj <> 0.0 Then
                    For i = j + 1 To n - 1
                        Dim xij As Double = lum(i)(j) / xjj
                        lum(i)(j) = xij
                        For k = j + 1 To n - 1
                            lum(i)(k) -= xij * lum(j)(k)
                        Next k
                    Next i
                End If
            Next j ' j

            Return toggle
        End Function ' MatrixDecompose

        Private Shared Function Helper(ByVal luMatrix()() As Double, ByVal b() As Double) As Double() ' helper
            Dim n As Integer = luMatrix.Length
            Dim x(n - 1) As Double
            b.CopyTo(x, 0)

            For i = 1 To n - 1
                Dim sum As Double = x(i)
                For j = 0 To i - 1
                    sum -= luMatrix(i)(j) * x(j)
                Next j
                x(i) = sum
            Next i

            x(n - 1) /= luMatrix(n - 1)(n - 1)
            For i = n - 2 To 0 Step -1
                Dim sum As Double = x(i)
                For j = i + 1 To n - 1
                    sum -= luMatrix(i)(j) * x(j)
                Next j
                x(i) = sum / luMatrix(i)(i)
            Next i

            Return x
        End Function ' Helper

        Private Shared Function MatrixDeterminant(ByVal matrix()() As Double) As Double
            Dim lum()() As Double = Nothing
            Dim perm() As Integer = Nothing
            Dim toggle As Integer = MatrixDecompose(matrix, lum, perm)
            Dim result As Double = toggle
            For i = 0 To lum.Length - 1
                result *= lum(i)(i)
            Next i
            Return result
        End Function

        ' ----------------------------------------------------------------

        Private Shared Function MatrixCreate(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            Dim result(rows)() As Double
            For i = 0 To rows - 1
                result(i) = New Double(cols) {}
            Next i
            Return result
        End Function

        Private Shared Function MatrixProduct(ByVal matrixA()() As Double, ByVal matrixB()() As Double) As Double()()
            Dim aRows As Integer = matrixA.Length
            Dim aCols As Integer = matrixA(0).Length
            Dim bRows As Integer = matrixB.Length
            Dim bCols As Integer = matrixB(0).Length
            If aCols <> bRows Then
                Throw New Exception("Non-conformable matrices")
            End If

            Dim result()() As Double = MatrixCreate(aRows, bCols)

            For i = 0 To aRows - 1 ' each row of A
                For j = 0 To bCols - 1 ' each col of B
                    For k = 0 To aCols - 1 ' could use k < bRows
                        result(i)(j) += matrixA(i)(k) * matrixB(k)(j)
                    Next k
                Next j
            Next i

            Return result
        End Function

        Private Shared Function MatrixAsString(ByVal matrix()() As Double) As String
            Dim s As String = ""
            For i = 0 To matrix.Length - 1
                For j = 0 To matrix(i).Length - 1
                    s &= matrix(i)(j).ToString("F3").PadLeft(8) & " "
                Next j
                s &= Environment.NewLine
            Next i
            Return s
        End Function

        Private Shared Function ExtractLower(ByVal lum()() As Double) As Double()()
            ' lower part of an LU Doolittle decomposition (dummy 1.0s on diagonal, 0.0s above)
            Dim n As Integer = lum.Length
            Dim result()() As Double = MatrixCreate(n, n)
            For i = 0 To n - 1
                For j = 0 To n - 1
                    If i = j Then
                        result(i)(j) = 1.0
                    ElseIf i > j Then
                        result(i)(j) = lum(i)(j)
                    End If
                Next j
            Next i
            Return result
        End Function

        Private Shared Function ExtractUpper(ByVal lum()() As Double) As Double()()
            ' upper part of an LU (lu values on diagional and above, 0.0s below)
            Dim n As Integer = lum.Length
            Dim result()() As Double = MatrixCreate(n, n)
            For i = 0 To n - 1
                For j = 0 To n - 1
                    If i <= j Then
                        result(i)(j) = lum(i)(j)
                    End If
                Next j
            Next i
            Return result
        End Function

        ' ----------------------------------------------------------------
    End Class ' Program
End Namespace ' ns
